package worker

import (
	"bytes"
	"context"
	"fmt"
	"io"

	"github.com/docker/docker/api/types"
	"github.com/docker/docker/api/types/container"
	"github.com/docker/docker/api/types/network"
	"github.com/docker/docker/client"
	"github.com/docker/go-units"
	"github.com/gliderlabs/ssh"
	log "github.com/noxiouz/zapctx/ctxlog"
	"github.com/rcrowley/go-metrics"
	"github.com/sonm-io/core/insonmnia/worker/plugin"
	"github.com/sonm-io/core/util/multierror"
	"go.uber.org/zap"
)

// NetworkStatsExt extends the Docker "NetworkStats" structure by adding
// rate meters.
type NetworkStatsExt struct {
	types.NetworkStats

	RxBytesRate   metrics.Meter
	RxPacketsRate metrics.Meter
	RxErrorsRate  metrics.Meter
	RxDroppedRate metrics.Meter
	TxBytesRate   metrics.Meter
	TxPacketsRate metrics.Meter
	TxErrorsRate  metrics.Meter
	TxDroppedRate metrics.Meter
}

// NewNetworkStatsExt constructs a new extended network stats, initializing
// with the given argument.
func NewNetworkStatsExt(v types.NetworkStats) *NetworkStatsExt {
	return &NetworkStatsExt{
		NetworkStats: v,

		RxBytesRate:   metrics.NewMeter(),
		RxPacketsRate: metrics.NewMeter(),
		RxErrorsRate:  metrics.NewMeter(),
		RxDroppedRate: metrics.NewMeter(),
		TxBytesRate:   metrics.NewMeter(),
		TxPacketsRate: metrics.NewMeter(),
		TxErrorsRate:  metrics.NewMeter(),
		TxDroppedRate: metrics.NewMeter(),
	}
}

// Update updates the current extended network stats, calculating rates.
func (m *NetworkStatsExt) Update(v types.NetworkStats) {
	m.RxBytesRate.Mark(int64(v.RxBytes - m.RxBytes))
	m.RxPacketsRate.Mark(int64(v.RxPackets - m.RxPackets))
	m.RxErrorsRate.Mark(int64(v.RxErrors - m.RxErrors))
	m.RxDroppedRate.Mark(int64(v.RxDropped - m.RxDropped))
	m.TxBytesRate.Mark(int64(v.TxBytes - m.TxBytes))
	m.TxPacketsRate.Mark(int64(v.TxPackets - m.TxPackets))
	m.TxErrorsRate.Mark(int64(v.TxErrors - m.TxErrors))
	m.TxDroppedRate.Mark(int64(v.TxDropped - m.TxDropped))

	m.NetworkStats = v
}

type containerStats struct {
	types.StatsJSON

	NetworksExt map[string]*NetworkStatsExt
}

func newContainerStats() *containerStats {
	return &containerStats{
		NetworksExt: map[string]*NetworkStatsExt{},
	}
}

func (m *containerStats) Update(v types.StatsJSON) {
	for name, networkStats := range v.Networks {
		networkStatsExt, ok := m.NetworksExt[name]
		if !ok {
			networkStatsExt = NewNetworkStatsExt(networkStats)
			m.NetworksExt[name] = networkStatsExt
		}

		networkStatsExt.Update(networkStats)
	}

	m.StatsJSON = v
}

type containerDescriptor struct {
	client *client.Client
	log    *zap.SugaredLogger

	ID              string
	CommitedImageID string
	description     Description
	stats           *containerStats

	cleanup plugin.Cleanup
}

func attachContainer(ctx context.Context, dockerClient *client.Client, ID string, d Description, tuners *plugin.Repository) (*containerDescriptor, error) {
	log.S(ctx).Infof("attaching to container with ID: %s reference: %s", ID, d.Reference.String())

	cont := containerDescriptor{
		ID:          ID,
		client:      dockerClient,
		description: d,
		stats:       newContainerStats(),
	}

	cleanup, err := tuners.GetCleanup(ctx, &d)
	if err != nil {
		return nil, err
	}

	cont.cleanup = cleanup
	cont.log = log.S(ctx).With(zap.String("container_id", cont.ID))

	return &cont, nil
}

func newContainer(ctx context.Context, dockerClient *client.Client, d Description, tuners *plugin.Repository) (*containerDescriptor, error) {
	log.S(ctx).Infof("start container with application, reference %s", d.Reference.String())

	cont := containerDescriptor{
		client:      dockerClient,
		description: d,
		stats:       newContainerStats(),
	}

	exposedPorts, portBindings, err := d.Expose()
	if err != nil {
		log.G(ctx).Error("failed to parse `expose` section", zap.Error(err))
		return nil, err
	}

	log.G(ctx).Debug("exposing ports", zap.Any("portBindings", portBindings))

	// NOTE: command to launch must be specified via ENTRYPOINT and CMD in Dockerfile
	var config = container.Config{
		AttachStdin:  false,
		AttachStdout: false,
		AttachStderr: false,

		ExposedPorts: exposedPorts,

		Image: d.Reference.String(),
		// TODO: set actual name
		Labels:  map[string]string{overseerTag: "", dealIDTag: d.DealId},
		Env:     d.FormatEnv(),
		Volumes: make(map[string]struct{}),
	}

	// NOTE: all ports are EXPOSE as PublishAll
	// TODO: detect network network mode and interface
	logOpts := make(map[string]string)
	// TODO: Move to StartTask?
	logOpts["max-size"] = "100m"
	var hostConfig = container.HostConfig{
		LogConfig:       container.LogConfig{Type: "json-file", Config: logOpts},
		PublishAllPorts: true,
		PortBindings:    portBindings,
		RestartPolicy:   d.RestartPolicy.Unwrap(),
		AutoRemove:      d.Autoremove,
		Resources:       d.Resources.ToHostConfigResources(d.CGroupParent),
	}

	networkingConfig := network.NetworkingConfig{
		EndpointsConfig: map[string]*network.EndpointSettings{},
	}
	if d.NetworkOptions != nil {
		networkingConfig.EndpointsConfig[d.NetworkOptions.Name] = &network.EndpointSettings{
			NetworkID: d.NetworkOptions.ID,
		}
	}

	additionalNetworking := &network.NetworkingConfig{}
	cleanup, err := tuners.Tune(ctx, &d, &hostConfig, additionalNetworking)
	if err != nil {
		log.G(ctx).Error("failed to tune container", zap.Error(err))
		return nil, err
	}
	log.G(ctx).Debug("successfully tuned container")

	log.G(ctx).Info("config", zap.Any("net", networkingConfig), zap.Any("host", hostConfig))
	// create new container
	// assign resulted containerid
	// log all warnings
	resp, err := cont.client.ContainerCreate(ctx, &config, &hostConfig, &networkingConfig, "")
	if err != nil {
		return nil, err
	}

	cleanup, err = tuners.PostCreationTune(ctx, &d, cleanup, resp.ID)
	if err != nil {
		return nil, err
	}

	for net, endpointCfg := range additionalNetworking.EndpointsConfig {
		log.S(ctx).Infof("connecting to %s network", net)
		if err := cont.client.NetworkConnect(ctx, endpointCfg.NetworkID, resp.ID, endpointCfg); err != nil {
			return nil, err
		}
	}

	cont.ID = resp.ID
	cont.log = log.S(ctx).With(zap.String("container_id", cont.ID))
	cont.cleanup = cleanup
	if len(resp.Warnings) > 0 {
		log.G(ctx).Warn("ContainerCreate finished with warnings", zap.Strings("warnings", resp.Warnings))
	}

	return &cont, nil
}

func (c *containerDescriptor) startContainer(ctx context.Context) error {
	var options types.ContainerStartOptions
	if err := c.client.ContainerStart(ctx, c.ID, options); err != nil {
		c.log.Warn("ContainerStart finished with error", zap.Error(err))
		return err
	}
	return nil
}

func (c *containerDescriptor) execCommand(ctx context.Context, cmd []string, env []string, isTty bool, wCh <-chan ssh.Window) (conn types.HijackedResponse, err error) {
	cfg := types.ExecConfig{
		User:         "root",
		Tty:          isTty,
		AttachStderr: true,
		AttachStdout: true,
		AttachStdin:  true,
		Detach:       false,
		Cmd:          cmd,
		Env:          env,
	}

	c.log.With(zap.Any("config", cfg)).Info("attaching command")

	execId, err := c.client.ContainerExecCreate(ctx, c.ID, cfg)
	if err != nil {
		c.log.Warnf("ContainerExecCreate finished with error: %s", err)
		return
	}

	conn, err = c.client.ContainerExecAttach(ctx, execId.ID, cfg)
	if err != nil {
		c.log.Warn("ContainerExecAttach finished with error: %s", err)
	}

	err = c.client.ContainerExecStart(ctx, execId.ID, types.ExecStartCheck{Detach: false, Tty: true})
	if err != nil {
		c.log.Warn("ContainerExecStart finished with error: %s", err)
		return
	}
	go func() {
		for {
			select {
			case w, ok := <-wCh:
				if !ok {
					return
				}
				c.log.Infof("resizing tty to %dx%d", w.Height, w.Width)
				err = c.client.ContainerExecResize(ctx, execId.ID, types.ResizeOptions{Height: uint(w.Height), Width: uint(w.Width)})
				if err != nil {
					log.G(ctx).Warn("ContainerExecResize finished with error", zap.Error(err))
				}
			case <-ctx.Done():
				return
			}
		}
	}()

	c.log.Info("attached command to container")
	return
}

func (c *containerDescriptor) Kill(ctx context.Context) error {
	c.log.Info("kill the container")
	if err := c.client.ContainerKill(ctx, c.ID, "SIGKILL"); err != nil {
		c.log.Warnf("failed to send SIGKILL to the container: %s", err)
		return err
	}
	return nil
}

func (c *containerDescriptor) Remove(ctx context.Context) error {
	c.log.Info("remove the container")
	result := multierror.NewMultiError()
	if err := containerRemove(ctx, c.client, c.ID); err != nil {
		result = multierror.Append(result, err)
	}
	if len(c.CommitedImageID) != 0 {
		if err := imageRemove(ctx, c.client, c.CommitedImageID); err != nil {
			result = multierror.Append(result, err)
		}
	}
	return result.ErrorOrNil()
}

//TODO: pass context
func (c *containerDescriptor) Cleanup() error {
	return c.cleanup.Close()
}

func containerRemove(ctx context.Context, client client.APIClient, id string) error {
	removeOpts := types.ContainerRemoveOptions{}
	if err := client.ContainerRemove(ctx, id, removeOpts); err != nil {
		return fmt.Errorf("failed to remove the container %s: %s", id, err)
	}
	log.S(ctx).Infof("removed container %s", id)
	return nil
}

func imageRemove(ctx context.Context, client client.APIClient, id string) error {
	removeOpts := types.ImageRemoveOptions{}
	if _, err := client.ImageRemove(ctx, id, removeOpts); err != nil {
		return fmt.Errorf("failed to remove the committed image %s: %s", id, err)
	}
	log.S(ctx).Infof("removed image %s", id)
	return nil
}

func (c *containerDescriptor) upload(ctx context.Context) error {
	if len(c.CommitedImageID) == 0 {
		opts := types.ContainerCommitOptions{}
		resp, err := c.client.ContainerCommit(ctx, c.ID, opts)
		if err != nil {
			return err
		}
		c.CommitedImageID = resp.ID
		c.log.Infof("committed container with new id %s", resp.ID)
	}
	tag := fmt.Sprintf("%s_%s", c.description.DealId, c.description.TaskId)

	newImg, err := c.description.Reference.WithTag(tag)
	if err != nil {
		c.log.Errorf("failed to add tag: %s", err)
		return err
	}

	c.log.Infof("tagging image %s from %s", newImg.String(), c.CommitedImageID)
	err = c.client.ImageTag(ctx, c.CommitedImageID, newImg.String())
	if err != nil {
		c.log.Errorf("failed to tag image: %s", err)
		return err
	}

	if c.description.PushOnStop {
		options := types.ImagePushOptions{
			RegistryAuth: c.description.Auth,
		}

		c.log.Infof("pushing image %s", newImg)
		reader, err := c.client.ImagePush(ctx, newImg.String(), options)
		if err != nil {
			c.log.Errorf("failed to push image: %s", err)
			return err
		}
		defer reader.Close()
		buf := bytes.NewBuffer(nil)
		if _, err := io.CopyN(buf, reader, units.MB); err != nil && err != io.EOF {
			return fmt.Errorf("pushing image: failed to read Docker's response: %v", err)
		}
		c.log.Infof("pushed image: %s", buf.String())
	}

	return nil
}
